import cv2
import numpy as np
from vhh_cmc.Configuration import Configuration
from vhh_cmc.PreProcessing import PreProcessing
import os
from vhh_cmc.OpticalFlow import OpticalFlow


#import matplotlib
#matplotlib.use('Qt5Agg')

class CMC(object):
    """
    Main class of camera movements classification (cmc) package.
    """
    def __init__(self, config_file: str):
        """
        Constructor

        :param config_file: [required] path to configuration file (e.g. PATH_TO/config.yaml)
                                       must be with extension ".yaml"
        """
        print("create instance of cmc ... ")

        if (config_file == ""):
            print("No configuration file specified!")
            exit()

        self.config_instance = Configuration(config_file)
        self.config_instance.loadConfig()

        if (self.config_instance.debug_flag == True):
            print("DEBUG MODE activated!")
            self.debug_results = self.config_instance.path_debug_results
            print("save debug results to: " + str(self.debug_results))

        self.pre_processing_instance = PreProcessing(config_instance=self.config_instance)

    def runOnSingleVideo(self, shots_per_vid_np=None, max_recall_id=-1):
        """
        Method to run cmc classification on specified video.

        :param shots_per_vid_np: [required] numpy array representing all detected shots in a video
                                 (e.g. sid | movie_name | start | end )
        :param max_recall_id: [required] integer value holding unique video id from VHH MMSI system
        """

        print("run cmc classifier on single video ... ")

        if (type(shots_per_vid_np) == None):
            print("ERROR: you have to set the parameter shots_per_vid_np!")
            exit()

        if (max_recall_id == -1 or max_recall_id == 0):
            print("ERROR: you have to set a valid max_recall_id [1-n]!")
            exit()

        if (self.config_instance.debug_flag == True):
            # load shot list from result file
            shots_np = self.loadSbdResults(self.config_instance.sbd_results_path)
        else:
            shots_np = shots_per_vid_np

        if (len(shots_np) == 0):
            print("ERROR: there must be at least one shot in the list!")
            exit()

        if (self.config_instance.debug_flag == True):
            num_shots = 1
        else:
            num_shots = len(shots_np)

        # read all frames of video
        vid_name = shots_np[0][0]

        if(self.config_instance.save_eval_results == 1):
            print("Evaluation mode is activated ...")
            cap = cv2.VideoCapture(vid_name)
        else:
            print(self.config_instance.path_videos + "/" + vid_name)
            cap = cv2.VideoCapture(self.config_instance.path_videos + "/" + vid_name)

        frame_l = []
        cnt = 0

        while(True):
            cnt = cnt + 1
            ret, frame = cap.read()
            # print(cnt)
            # print(ret)
            # print(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if (ret == True):
                frame = self.pre_processing_instance.applyTransformOnImg(frame)
                frame_l.append(frame)
                #cv2.imshow("asd", frame)
                #cv2.waitKey(1)
                #if(cnt == 100):
                #    exit()
            else:
                break

        all_frames_np = np.array(frame_l)
        #print(all_frames_np.shape)

        #all_frames_np = all_frames_np[1240:1478,:,:,:]
        shot_start_idx = 0  # used in debugging mode - to select specific shot
        results_cmc_l = []
        for idx in range(shot_start_idx, shot_start_idx + num_shots):
            #print(shots_np[idx])
            shot_id = int(shots_np[idx][1])
            vid_name = str(shots_np[idx][0])
            start = int(shots_np[idx][2])
            stop = int(shots_np[idx][3])
            shot_frames_np = all_frames_np[start:stop + 1, :, :, :]
            shot_len = stop - start

            MIN_NUMBER_OF_FRAMES_PER_SHOT = 10
            if(shot_len <= MIN_NUMBER_OF_FRAMES_PER_SHOT ):
                #print("shot length is too small!")
                class_name = "NA"
            else:
                # add new optical flow version
                optical_flow_instance = OpticalFlow(video_frames=shot_frames_np,
                                                    algorithm="orb",
                                                    config_instance=self.config_instance)
                mag_l, angles_l = optical_flow_instance.run()
                class_name = optical_flow_instance.predict_final_result(mag_l,
                                                                        angles_l,
                                                                        self.config_instance.class_names)


            '''
            # run optical flow process
            optical_flow_instance = OpticalFlow(video_frames=shot_frames_np,
                                                fPath=self.config_instance.path_videos + "/" + vid_name,
                                                debug_path=self.config_instance.path_raw_results,
                                                sf=start,
                                                ef=stop,
                                                mode=self.config_instance.mode,
                                                sensitivity=self.config_instance.sensitivity,
                                                specificity=self.config_instance.specificity,
                                                border=self.config_instance.border,
                                                number_of_features=self.config_instance.number_of_features,
                                                angle_diff_limit=self.config_instance.angle_diff_limit,
                                                config=None)
            pan_list, tilt_list = optical_flow_instance.run()

            print(pan_list)
            print(tilt_list)

            number_of_all_frames = abs(start - stop)
            if number_of_all_frames == 0:
                number_of_all_frames = 0.000000000001

            number_of_pan_frames = 0
            for sf, ef in pan_list:
                diff = abs(sf - ef)
                number_of_pan_frames = number_of_pan_frames + diff
            pans_score = int((number_of_pan_frames * 100) / number_of_all_frames)
            print(pans_score)

            number_of_tilt_frames = 0
            for sf, ef in tilt_list:
                diff = abs(sf - ef)
                number_of_tilt_frames = number_of_tilt_frames + diff
            tilts_score = int((number_of_tilt_frames * 100) / number_of_all_frames)
            print(tilts_score)

            #if(self.config_instance.save_eval_results == 1):
            #    if (pans_score >= threshold):
            #        class_name = self.config_instance.class_names[0]
            #    else:
            #        class_name = self.config_instance.class_names[1]
            #else:
            threshold = 60
            if (pans_score >= threshold):
                class_name = self.config_instance.class_names[0]
            elif(tilts_score >= threshold):
                class_name = self.config_instance.class_names[1]
            elif (pans_score >= threshold) and (tilts_score >= threshold):
                class_name = self.config_instance.class_names[2]
            else:
                class_name = self.config_instance.class_names[2]
            '''

            # prepare results
            print(str(vid_name) + ";" + str(shot_id) + ";" + str(start) + ";" + str(stop) + ";" + str(class_name))
            results_cmc_l.append([str(vid_name) + ";" + str(shot_id) + ";" + str(start) + ";" + str(stop) + ";" + str(class_name)])
            #exit()

            '''
            # save raw results
            if(self.config_instance.save_raw_results == 1):
                optical_flow_instance.to_csv(
                    "/".join([self.config_instance.path_raw_results, self.config_instance.path_prefix_raw_results + str(vid_name.split('/')[-1]) + ".csv"]))
                optical_flow_instance.to_png(
                    "/".join([self.config_instance.path_raw_results, self.config_instance.path_prefix_raw_results + str(vid_name) + ".png"]))
                optical_flow_instance.to_avi(
                    "/".join([self.config_instance.path_raw_results, self.config_instance.path_prefix_raw_results + str(vid_name) + ".avi"]))
            '''
        results_cmc_np = np.array(results_cmc_l)
        print(results_cmc_np)
        #exit()

        # export results
        if (self.config_instance.save_eval_results == 1):
            self.exportCmcResults(vid_name, results_cmc_np)
        else:
            self.exportCmcResults(str(max_recall_id), results_cmc_np)

    def loadSbdResults(self, sbd_results_path):
        """
        Method for loading shot boundary detection results as numpy array

        .. note::
            Only used in debug_mode.

        :param sbd_results_path: [required] path to results file of shot boundary detection module (vhh_sbd)
        :return: numpy array holding list of detected shots.
        """

        # open sbd results
        fp = open(sbd_results_path, 'r')
        lines = fp.readlines()
        lines = lines[1:]

        lines_n = []
        for i in range(0, len(lines)):
            line = lines[i].replace('\n', '')
            line_split = line.split(';')
            lines_n.append([line_split[0], os.path.join(line_split[1]), line_split[2], line_split[3]])
        lines_np = np.array(lines_n)
        # print(lines_np.shape)

        return lines_np

    def exportCmcResults(self, fName, cmc_results_np: np.ndarray):
        """
        Method to export cmc results as csv file.

        :param fName: [required] name of result file.
        :param cmc_results_np: numpy array holding the camera movements classification predictions for each shot of a movie.
        """

        print("export results to csv!")

        if (len(cmc_results_np) == 0):
            print("ERROR: numpy is empty")
            exit()

        # open stc resutls file
        if (self.config_instance.debug_flag == True):
            fp = open(self.debug_results + "/" + fName + ".csv", 'w')
        elif (self.config_instance.save_eval_results == 1):

            print(self.config_instance.path_eval_results)
            fp = open(self.config_instance.path_eval_results + "/" + fName.split('/')[-1].split('.')[0] + ".csv", 'w')
        else:
            fp = open(self.config_instance.path_final_results + "/" + fName + ".csv", 'w')



        header = "vid_name;shot_id;start;end;cmc"
        fp.write(header + "\n")

        for i in range(0, len(cmc_results_np)):
            tmp_line = str(cmc_results_np[i][0])
            for c in range(1, len(cmc_results_np[i])):
                tmp_line = tmp_line + ";" + cmc_results_np[i][c]
            fp.write(tmp_line + "\n")

        fp.close()



import cv2
import numpy as np
import argparse
from matplotlib import pyplot as plt
from vhh_cmc.OpticalFlow_ORB import OpticalFlow_ORB
from vhh_cmc.OpticalFlow_SIFT import OpticalFlow_SIFT

class OpticalFlow(object):
    def __init__(self, video_frames=None, algorithm="orb", config_instance=None):
        #self.video_name = "C:\\Users\\dhelm\\Documents\\slow_traffic_small.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\tilt\\b88f0e71-a0f2-4efe-ae0d-5b83a0770b73_32.mp4"  # tilt unten nach oben
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\tilt\\35_25178.mp4"  # tilt oben nach unten
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\tilt\\94_24357.mp4" # tilt unten nach oben
        #1371a561-6b19-4d69-8210-1347ca75eb90_96
        # self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\pan\\1371a561-6b19-4d69-8210-1347ca75eb90_104.mp4"
        # self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\tilt\\tilt_130_74088.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\pan\\130_74173.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\pan\\1371a561-6b19-4d69-8210-1347ca75eb90_96.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\pan\\066f929d-6434-4ea9-844b-e066f57b6c28_99.mp4" # rechts nach links
        #self.video_name = "C:\\Users\\dhelm\\Documents\\999.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\444.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\pan\\72_65942.mp4"  # pan links nach rechts

        self.video_frames = video_frames
        self.config_instance = config_instance

        if (algorithm == "sift"):
            self.feature_detector = OpticalFlow_SIFT(video_frames=video_frames)
        elif (algorithm == "orb"):
            self.feature_detector = OpticalFlow_ORB(video_frames=video_frames)
        else:
            print("ERROR: select valid feature extractor [e.g. sift, orb, pesc]")
            exit()

    def run(self):

        '''
        frame_list = []
        cap = cv2.VideoCapture(self.video_name)
        while(1):
            ret, old_frame = cap.read()
            if(ret == False):
                break
            old_gray = cv2.cvtColor(old_frame, cv2.COLOR_BGR2GRAY)
            old_gray = self.crop(old_gray, (500, 500))
            frame_list.append(old_gray)
        frames_np = np.array(frame_list)
        print(frames_np.shape)
        cap.release()
        '''

        frames_np = self.video_frames

        filtered_mag_l_n = []
        filtered_angles_l_n = []

        angles_l_n = []
        mag_l_n = []
        for i in range(1, len(frames_np)):
            #print("##########")

            prev_frame = frames_np[i - 1]
            curr_frame = frames_np[i]

            distance_threshold = self.config_instance.distance_threshold
            kp_prev_list, kp_curr_list = self.feature_detector.getMatches(prev_frame, curr_frame, distance_threshold)

            #print("---")
            #print("number of features")
            #print(len(kp_curr_list))
            #print(len(kp_prev_list))

            if (len(kp_prev_list) == 0 or len(kp_curr_list) == 0):
                mag_l_n.append([0, 0])
                angles_l_n.append([0, 0])
                continue

            curr_points = np.array(kp_curr_list).astype('float').reshape(-1, 1, 2)
            prev_points = np.array(kp_prev_list).astype('float').reshape(-1, 1, 2)
            mag_n, angle_n = self.compute_magnitude_angle(prev_points,
                                                          curr_points)
            # angle_raw.append(angle_n.tolist())
            # mag_raw.append(mag_n.tolist())

            mag_n = np.abs(mag_n)  # [:50])
            #mag_n, outliers_idx = self.filter1D(mag_n, alpha=3)
            #mag_n = np.delete(mag_n, outliers_idx)

            mag_mean_n = np.mean(mag_n)
            mag_l_n.append([0, mag_mean_n])

            angle_n = np.abs(angle_n)  # [:50])
            angle_mean_n = np.mean(angle_n)
            angles_l_n.append([0, angle_mean_n])

            filtered_angle_n = angles_l_n
            filtered_angles_l_n.append(filtered_angle_n)

            '''
            data_std = np.std(mag_n)
            data_mean = np.mean(mag_n)
            anomaly_cut_off = data_std * 3

            lower_limit = data_mean - anomaly_cut_off
            upper_limit = data_mean + anomaly_cut_off
            print(lower_limit)
            # Generate outliers
            outliers_idx = []
            for o, outlier in enumerate(mag_n):
                if outlier > upper_limit or outlier < lower_limit:
                    outliers_idx.append(o)
            filtered_mag_n = np.delete(mag_n, outliers_idx)
            filtered_angle_n = np.delete(angle_n, outliers_idx)
            filtered_curr_points = np.delete(curr_points, outliers_idx, axis=0)
            filtered_prev_points = np.delete(prev_points, outliers_idx, axis=0)

            filtered_mag_n = np.abs(filtered_mag_n)  # [:50])
            filtered_mag_mean_n = np.median(filtered_mag_n)
            filtered_mag_l_n.append([0, filtered_mag_mean_n])

            filtered_angle_n = np.abs(filtered_angle_n)  # [:50])
            filtered_angle_mean_n = np.median(filtered_angle_n)
            '''

            '''
            # plot angles over time
            #plt.figure(1)
            fig, axs = plt.subplots(2)
            fig.suptitle('mag and angles per feature point in one frame')
            axs[0].plot(np.arange(len(mag_n)), mag_n)
            axs[0].plot(outliers_idx, mag_n[outliers_idx])
            axs[0].plot(np.arange(len(filtered_mag_n)), filtered_mag_n)
            axs[1].plot(np.arange(len(angle_n)), angle_n)
            #plt.ylim(ymax=190, ymin=-190)
            plt.grid(True)
            plt.show()
            #plt.draw()
            #plt.pause(0.02)
            '''

            '''
            # draw the tracks
            mask = np.zeros_like(frames_np[i])
            for j, (new, old) in enumerate(zip(curr_points, prev_points)):
                if(j > 10):
                    break;
                a, b = new.ravel().astype('int')
                c, d = old.ravel().astype('int')
                mask = cv2.line(mask, (a, b), (c, d), (255, 0, 0), 1)
                frame_curr = cv2.circle(frames_np[i], (a, b), 5, (255, 0, 0), -1)
            img = cv2.add(frame_curr, mask)
            cv2.imshow('frame', img)
            '''

            '''
            # draw orig with n feature points
            n_feature_points = len(curr_points)
            for j, (new, old) in enumerate(zip(curr_points, prev_points)):
                if (j > n_feature_points):
                    break
                a, b = new.astype('int').ravel()
                c, d = old.astype('int').ravel()
                frame_curr = cv2.circle(frames_np[i], (a, b), 2, (255, 0, 0), -1)
                frame_curr = cv2.line(frame_curr, (a, b), (a + 5, b + 5), (255, 0, 0), 1)
            # img = cv2.add(frame_curr, mask)
            cv2.imshow('frame', frame_curr)

            k = cv2.waitKey(30) & 0xff
            if k == 27:
                break
            '''

            '''
            #print(curr_points)
            #print(curr_points.shape)
            #print(filtered_curr_points)
            #print(filtered_curr_points.shape)
            # draw orig with n feature points
            n_feature_points = len(filtered_curr_points)
            for j, (new, old) in enumerate(zip(filtered_curr_points, filtered_prev_points)):
                if (j > n_feature_points):
                    break
                a, b = new.astype('int').ravel()
                c, d = old.astype('int').ravel()
                frame_curr = cv2.circle(frames_np[i], (a, b), 3, (255, 0, 0), -1)
                #frame_curr = cv2.line(frame_curr, (a, b), (a + int(vector_x[j]) * 1, b),
                #                      (0, 0, 255), 2)
                #frame_curr = cv2.line(frame_curr, (a, b), (a, b + int(vector_y[j]) * 1),
                #                      (0, 255, 0), 2)

            # img = cv2.add(frame_curr, mask)
            cv2.imshow('frame', frame_curr)

            k = cv2.waitKey(30) & 0xff
            if k == 27:
                break

            '''

            '''
            # plot angles over time
            #plt.figure(1)
            fig, axs = plt.subplots(3)
            fig.suptitle('mag and angles per feature point in one frame')
            axs[0].plot(np.arange(len(mag_n)), mag_n)
            axs[0].plot(outliers_idx, mag_n[outliers_idx])
            axs[0].plot(np.arange(len(filtered_mag_n)), filtered_mag_n)
            axs[1].plot(np.arange(len(angle_n)), angle_n)
            axs[2].imshow(frame_curr)
            #plt.ylim(ymax=190, ymin=-190)
            plt.grid(True)
            plt.show()
            #plt.draw()
            #plt.pause(0.02)
            '''

            '''
            # plot angles over time
            #plt.figure(1)
            fig, axs = plt.subplots(2)
            fig.suptitle('mag and angles per feature point in one frame')
            axs[0].plot(np.arange(len(mag_n)), mag_n)
            axs[0].plot(outliers_idx, mag_n[outliers_idx])
            axs[0].plot(np.arange(len(filtered_mag_n)), filtered_mag_n)
            axs[1].plot(np.arange(len(angle_n)), angle_n)
            #plt.ylim(ymax=190, ymin=-190)
            plt.grid(True)
            plt.show()
            #plt.draw()
            #plt.pause(0.02)
            '''

        # cv2.destroyAllWindows()

        return mag_l_n, angles_l_n

    def predict_final_result(self, mag_l_n, angles_l_n, class_names):
        # print(type(mag_l_n))
        # print(len(mag_l_n))
        # exit()

        # calcualate final result
        angles_np = np.array(angles_l_n)
        mag_np = np.array(mag_l_n)

        # add filter
        #print(mag_np[:, 1:])
        filtered_mag_n, outlier_idx = self.filter1D(mag_np[:, 1:], alpha=3)
        filtered_angles_np = np.delete(angles_np[:, 1:], outlier_idx)

        filtered_angle_n, outlier_idx = self.filter1D(filtered_angles_np, alpha=2)
        filtered_mag_n = np.delete(filtered_mag_n, outlier_idx)

        # calculate x - y components - NOT USED YET
        vector_y = np.multiply(filtered_mag_n, np.sin(np.deg2rad(filtered_angle_n)))
        vector_x = np.multiply(filtered_mag_n, np.cos(np.deg2rad(filtered_angle_n)))

        b, bins, patches = plt.hist(filtered_angle_n, bins=8, range=[0, 360],
                                    cumulative=False)  # bins=None, range=None

        th = self.config_instance.min_magnitude_threshold  # 2.0  # manual set threshold for magnitude

        DEBUG_VIS_FLAG = False
        if(DEBUG_VIS_FLAG == True):
            # plot angles over time (frames)
            fig, axs = plt.subplots(3)
            fig.suptitle('mag and angles per feature point in one frame')
            a = np.empty(len(filtered_mag_n))
            a.fill(th)
            axs[0].plot(np.arange(len(filtered_mag_n)), a)
            a = np.empty(len(filtered_mag_n))
            a.fill(np.median(filtered_mag_n))
            axs[0].plot(np.arange(len(filtered_mag_n)), a)
            np.median(filtered_mag_n)
            axs[0].plot(np.arange(len(filtered_mag_n)), filtered_mag_n)
            axs[1].plot(np.arange(len(filtered_angle_n)), filtered_angle_n)
            b, bins, patches = axs[2].hist(filtered_angle_n, bins=8, range=[0,360], cumulative=False)  #bins=None, range=None
            #plt.ylim(ymax=190, ymin=-190)
            plt.grid(True)
            plt.show()
        ''''''
        ''''''

        percentage = 0.5  # ratio threshold between no-movement and movement
        class_names_n = ['PAN', 'TILT', 'TILT', 'PAN', 'PAN', 'TILT', 'TILT', 'PAN']

        print("predicted median magnitude: " + str(np.median(filtered_mag_n)))
        print("threshold magnitude: " + str(th))

        class_name = class_names_n[np.argmax(b)]
        if ((class_name == "PAN" and np.median(filtered_mag_n) > th)):
            class_name = "PAN"
        elif ((class_name == "TILT" and np.median(filtered_mag_n) > th)):
            class_name = "TILT"
        else:
            class_name = "NA"

        '''
        print(np.mean(angles_np))
        angle = np.mean(angles_np)
        print("predicted angle: " + str(angle))

        print(np.mean(mag_np))
        mag = np.mean(mag_np)
        print("predicted mag: " + str(mag))

        if ((angle >= 140 and angle < 220) or (angle >= -40 and angle < 40) and (mag > 20)):
            class_name = class_names[0]
        elif ((angle >= 50 and angle < 130) or (angle >= 230 and angle < 310) and (mag > 20)):
            class_name = class_names[1]
        else:
            class_name = class_names[2]
        print(class_name)
        '''
        return class_name

    def filter1D(self, data_np, alpha=2):
        print(type(data_np))
        print(data_np.shape)

        data_std = np.std(data_np)
        data_mean = np.mean(data_np)
        anomaly_cut_off = data_std * alpha

        lower_limit = data_mean - anomaly_cut_off
        upper_limit = data_mean + anomaly_cut_off
        print(lower_limit)
        # Generate outliers
        outliers_idx = []
        for o, outlier in enumerate(data_np):
            if outlier > upper_limit or outlier < lower_limit:
                outliers_idx.append(o)
        filtered_data_np = np.delete(data_np, outliers_idx)
        return filtered_data_np, outliers_idx

    def crop(self, img: np.ndarray, dim: tuple):
        """
        This method is used to crop a specified region of interest from a given image.

        :param img: This parameter must hold a valid numpy image.
        :param dim: This parameter must hold a valid tuple including the crop dimensions.
        :return: This method returns the cropped image.
        """
        crop_w, crop_h = dim

        crop_h_1 = 0
        crop_h_2 = 0
        crop_w_1 = 0
        crop_w_2 = 0

        img_h = img.shape[0]
        img_w = img.shape[1]

        crop_w_1 = int(img_w / 2) - int(crop_w / 2)
        if (crop_w_1 < 0):
            crop_w_1 = 0

        crop_w_2 = int(img_w / 2) + int(crop_w / 2)
        if (crop_w_2 >= img_w):
            crop_w_2 = img_w

        crop_h_1 = int(img_h / 2) - int(crop_h / 2)
        if (crop_h_1 < 0):
            crop_h_1 = 0

        crop_h_2 = int(img_h / 2) + int(crop_h / 2)
        if (crop_h_2 >= img_h):
            crop_h_2 = img_h

        img_crop = img[crop_h_1:crop_h_2, crop_w_1:crop_w_2]
        return img_crop

    def compute_magnitude_angle(self, prev_feat, curr_feat):
        if prev_feat.__len__() <= 0:
            print("no previous features... returning")
            assert (prev_feat.__len__() > 0)
        if prev_feat.__len__() != curr_feat.__len__():
            print("length is not correct")
            assert (prev_feat.__len__() == curr_feat.__len__())
        d = curr_feat - prev_feat
        #print(d.shape)
        mag = np.hypot(d[:, 0, 0], d[:, 0, 1])
        ang = np.round(np.degrees(np.arctan2(d[:, 0, 1], d[:, 0, 0])))

        # ang = np.round(np.arctan2(d[:, 0, 1], d[:, 0, 0])*180 / np.pi)

        return mag, ang



import cv2
import numpy as np
from enum import IntEnum
from matplotlib import pyplot as plt
from vhh_cmc import cmc_io
from vhh_cmc.Configuration import Configuration
import os


class ClassificationType(IntEnum):
    """
    Class holds enum values to specify PAN and TILT.
    """
    PAN = 0,
    TILT = 1


class AngleClassifier(object):
    """
    This class is used to classify different camera movements (e.g. Tilt and Pan).
    """

    # r ... range of angles allowed
    def __init__(self, classification_type, r=np.array([-25, 25])):
        """
        Constructor

        :param classification_type:
        :param r: This parameter specifies the range of valid angles (e.g. [-25, 25])
        """

        self.r = r
        self.classification_type = classification_type
        if self.classification_type == ClassificationType.PAN:
            self.main_angles = np.array([180, 0])
        elif self.classification_type == ClassificationType.TILT:
            self.main_angles = np.array([90, -90])

    def classify(self, angle):
        """
        This method is used to classify a given angle to one out of the classes PAN or TILT.

        :param angle: This parameter represents a angle (e.g. value in range [0, 360])
        :return: This method returns a boolean flag which is set to TRUE if the angle corresponds to a specified range of TILT or PAN.
        """

        print("Checking for " + self.classification_type.name + ": " + str(angle) + " degrees.")
        for i, a in enumerate(self.main_angles):
            print("Classifying: distance " + str(self.r) + " to angle " + str(a))
            if int(angle) in range(self.r[0]+a, self.r[1]+a):
                print("Angle contributes to " + self.classification_type.name + " movement.")
                return True
        print("Angle does not contribute to " + self.classification_type.name + "movement.")
        return False

    def __str__(self):
        """
        This method is used to get the camera movement type.

        :return: This method returns the name (string) of the camera movement (e.g. PAN or TILT).
        """
        return self.classification_type.name


class ClassificationCounter(object):
    # classification = [ frame index, is movement]
    # classification[i] = [ i, is movement ]
    # classification[:, 0] = [ ... , -i, ..., 0, ..., i, ... ]
    # classification[:, 1] = [ is movement, is movement, ..., is movement ... ]

    def inc(self, i):
        self.counter[i] = self.counter[i] + 1

    def reset(self, i):
        self.counter[i] = 0

    def fill_totals(self, i, v, o):
        self.totals[i - 1] = np.array([i + o, v])

    def oversteps(self, i, v):
        return True if self.counter[i] > v else False

    def equals(self, i, v):
        return True if self.counter[i] == v else False

    def get(self, i):
        return self.counter[i]

    def add_movement(self, i):
        self.movements.append(np.array([i, -1]))
        print("creating movement at start frame ", i)
        self.movement_created = True

    def end_movement(self, i):
        if self.movement_created:
            self.movements[self.movements.__len__() - 1][1] = i
            print("movement [" + str(self.movements[self.movements.__len__() - 1][0]) + ", " +
                  str(self.movements[self.movements.__len__() - 1][0]) + "]")
            self.movement_created = False

    def resize(self, size, i):
        self.totals = np.resize(self.totals, [size, 2])
        if self.movement_created:
            self.end_movement(i)

    def is_movement(self, frame_idx):
        is_movement = np.array([(sf <= frame_idx <= ef) for (sf, ef) in self.movements])
        return np.any(is_movement)

    def draw_movement(self, xy, o, c='red', c2='darkred', l='movement'):
        points = xy[:, self.totals[:, 1].astype(bool)]
        plt.scatter(points[0], points[1], color=c, label=l)

        xmin = np.array([sf for (sf, ef) in self.movements])
        xmax = np.array([ef for (sf, ef) in self.movements])

        if xmin.__len__() == 0:
            return
        ymin = xy[1, xmin - o]
        ymax = xy[1, xmax - o]

        yavg = ymin# + ymax) / 2.
        plt.hlines(yavg, xmin, xmax, colors=c2, linewidth=5)
        plt.vlines(xmin, yavg - 10, yavg + 10, colors=c2, linewidth=5)
        plt.vlines(xmax, yavg - 10, yavg + 10, colors=c2, linewidth=5)

    def __init__(self, totals):
        self.counter = np.zeros(2).astype(int)
        self.totals = totals
        # assert at least one fake movement.
        self.movements = []
        self.movement_created = False


class Runmodi(IntEnum):
    """
    This class holds enum values to specify runtime mode.
    """
    NORMAL_MODE = 0
    DEBUG_MODE = 1
    SAVE_MODE = 2
    DEBUG_AND_SAVE_MODE = 3


class OpticalFlow_Pesc(object):
    """
    This class is used for optical flow calculation.
    """

    def __init__(self,
                 video_frames=None,
                 fPath="",
                 debug_path="",
                 sf=0,
                 ef=1,
                 mode=0,
                 pan_classifier=AngleClassifier(ClassificationType.PAN),
                 tilt_classifier=AngleClassifier(ClassificationType.TILT),
                 sensitivity=20,
                 specificity=3,
                 border=50,
                 number_of_features=100,
                 angle_diff_limit=20,
                 config=None):
        """
        Constructor.

        :param video_frames: This parameter holds a valid numpy array representing a range of frames (e.g. NxWxHxchannels).
        :param fPath: This parameter holds a valid path consisting of the absolute path and the correct video name.
        :param debug_path: This parameter specifies a valid path to store results in debug mode.
        :param sf: This parameter represents the starting frame index.
        :param ef: This parameter represents the ending frame index.
        :param mode: This parameter represents runtime mode (e.g. DEBUG_MODE=1 or SAVE_MODE=2).
        :param pan_classifier: This parameter holds a valid object of class type AngleClassifier.
        :param tilt_classifier: This parameter holds a valid object of class type AngleClassifier.
        :param sensitivity: This parameter is used to configure the optical flow algorithm.
        :param specificity: This parameter is used to configure the optical flow algorithm.
        :param border: This parameter is used to configure the optical flow algorithm.
        :param number_of_features: This parameter is used to configure the optical flow algorithm.
        :param angle_diff_limit: This parameter is used to configure the optical flow algorithm.
        :param config:
        """
        self.debug_path = debug_path
        self.video_frames = video_frames
        self.fpath = fPath
        self.sf = sf
        self.ef = ef
        self.mode = Runmodi(mode)
        self.sensitivity = sensitivity
        self.specificity = specificity
        self.border = border
        self.number_of_features = number_of_features
        self.angle_diff_limit = angle_diff_limit

        if config is not None:
            self.configure(config)

        self.i = 1
        self.end = self.ef - self.sf
        if(self.end == 0): self.end = 1

        self.most_common_angles = np.zeros([self.end - 1, 1])
        self.weights = np.zeros([self.end - 1, 1])
        self.frame_size = (0, 0)

        self.cap = None

        self.out = {}

        self.pan_classifier = pan_classifier
        self.tilt_classifier = tilt_classifier
        self.pan_counter = ClassificationCounter(np.zeros([self.end, 2]).astype(int))
        self.tilt_counter = ClassificationCounter(np.zeros([self.end, 2]).astype(int))
        # lists containing all movements: [ start frame, end frame ]
        self.pans = []
        self.tilts = []

        # key handling
        self.key_handler = cmc_io.KeyHandler(50, {
            27: self.interrupt,
            ord('s'): self.init_savemode,
            ord('d'): self.init_debugmode,
            ord('b'): self.init_savedebugmode,
            ord('n'): self.init_normalmode,
            ord('p'): self.annotate_pan,
            ord('t'): self.annotate_tilt,
            ord('e'): self.annotate_nomove
        })

        print("Creating new OpticalFlowCameraMovementClassifier " + str(self))

    # run optical flow computation
    def run(self):
        """
        This method is used to run the optical flow calculation process.

        :return: This method returns a separate list for each movement class and holds the predicted frame ranges of both.
        """
        curr_frame = self.video_frames[0]
        h, w, _ = curr_frame.shape
        self.frame_size = (w, h)

        if self.mode > Runmodi.DEBUG_MODE:
            # need to compute output path
            print("Running program in save mode: configuring outputs.")
            self.out = self.init_outputs(self.fpath, self.sf, self.ef, (w, h))

        # TODO: replace with meaningful feature extractor
        #curr_feat = self.create_random_features(self.number_of_features, self.range_of_frame(curr_frame))
        #print(len(curr_feat))

        # params for ShiTomasi corner detection
        feature_params = dict(maxCorners=500,
                              qualityLevel=0.000003,
                              minDistance=5,
                              blockSize=128)

        print(curr_frame.shape)
        curr_frame_gray = cv2.cvtColor(curr_frame, cv2.COLOR_BGR2GRAY)
        curr_feat = cv2.goodFeaturesToTrack(curr_frame_gray, mask=None, **feature_params)

        #prev_feat, curr_feat = self.getORBMatches(prev_frame_gray, curr_frame_gray)

        #print(len(curr_feat))
        #exit()
        raw_angles_l = []
        raw_mag_l = []
        while self.i < self.end:
            prev_frame, prev_feat, curr_frame, curr_feat = self.step(curr_frame, curr_feat)

            print("prev: ", prev_feat.__len__())
            print("curr: ", curr_feat.__len__())

            prev_feat, curr_feat = self.optical_flow(prev_frame, prev_feat, curr_frame)
            mag, angle = self.compute_magnitude_angle(prev_feat, curr_feat)

            mag_mean = np.mean(mag)
            angle_mean = np.mean(angle)
            raw_angles_l.append(angle_mean)
            raw_mag_l.append(mag_mean)


            most_common_angle = None
            weight = 1
            if self.i >= 2:
                most_common_angle = self.most_common_angles[self.i - 2]
                weight = self.weights[self.i - 2]

                # classifiy movement
                #self.check_for_movement(most_common_angle, self.pan_counter, self.pan_classifier)
                #self.check_for_movement(most_common_angle, self.tilt_counter, self.tilt_classifier)
                print("----")
                print(most_common_angle)
                print(angle_mean)
                self.check_for_movement(angle_mean, self.pan_counter, self.pan_classifier)
                self.check_for_movement(angle_mean, self.tilt_counter, self.tilt_classifier)

            print("prev: ", prev_feat.__len__())
            print("curr: ", curr_feat.__len__())

            if prev_feat.__len__() == 0:
                self.most_common_angles[self.i - 1] = self.most_common_angles[self.i - 2]
                self.weights[self.i - 1] = self.weights[self.i - 2]
            else:
                self.most_common_angles[self.i - 1], curr_feat, self.weights[self.i - 1] = \
                    self.estimate_background(prev_feat, curr_feat, most_common_angle, weight, curr_frame)
            self.i = self.i + 1

        self.pans = self.pan_counter.movements
        self.tilts = self.tilt_counter.movements

        for (sf, ef) in self.pan_counter.movements:
            print(sf)
            print(ef)
        for (sf, ef) in self.tilt_counter.movements:
            print(sf)
            print(ef)
        ''''''
        print(self.fpath)
        print("#############")
        print(raw_angles_l)
        print(raw_mag_l)

        '''
        # plot angles over time
        plt.figure()
        plt.plot(np.arange(len(raw_angles_l)), raw_angles_l)
        # plt.plot(np.arange(len(mag_l)), mag_l)
        plt.ylim(ymax=190, ymin=-190)
        plt.grid(True)
        # plt.show()
        plt.savefig("./raw_angles_" + str(self.fpath.split('/')[-1].split('.')[0]) + ".png")
        '''
        self.clear()
        return self.pans, self.tilts

    # run optical flow computation
    def run_eval(self):
        """
        This method is used to run the optical flow calculation to evaluate specified videos.
        """
        self.cap = cv2.VideoCapture(self.fpath)
        if self.ef == 1:
            self.re_init_ef(int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT)))
        self.cap.set(1, self.sf)

        print("\n".join(["Video path: " + self.fpath,
                         "first frame to be captured: " + str(self.sf),
                         "last frame to be captured: " + str(self.ef)]))

        ret, curr_frame = self.cap.read()

        if not ret:
            raise Exception("Failing to open video file.")

        h, w, _ = curr_frame.shape
        self.frame_size = (w, h)

        if self.mode > Runmodi.DEBUG_MODE:
            # need to compute output path
            print("Running program in save mode: configuring outputs.")
            self.out = self.init_outputs(self.fpath, self.sf, self.ef, (w, h))

        curr_feat = self.create_random_features(self.number_of_features, self.range_of_frame(curr_frame))

        while self.i < self.end and self.cap.isOpened():
            prev_frame, prev_feat, curr_frame, curr_feat = self.step_eval(curr_frame, curr_feat)

            print("prev: ", prev_feat.__len__())
            print("curr: ", curr_feat.__len__())

            prev_feat, curr_feat = self.optical_flow(prev_frame, prev_feat, curr_frame)

            most_common_angle = None
            weight = 1
            if self.i >= 2:
                most_common_angle = self.most_common_angles[self.i - 2]
                weight = self.weights[self.i - 2]

                # classifiy movement
                self.check_for_movement(most_common_angle, self.pan_counter, self.pan_classifier)
                self.check_for_movement(most_common_angle, self.tilt_counter, self.tilt_classifier)

            print("prev: ", prev_feat.__len__())
            print("curr: ", curr_feat.__len__())

            if prev_feat.__len__() == 0:
                self.most_common_angles[self.i - 1] = self.most_common_angles[self.i - 2]
                self.weights[self.i - 1] = self.weights[self.i - 2]
            else:
                self.most_common_angles[self.i - 1], curr_feat, self.weights[self.i - 1] = \
                    self.estimate_background(prev_feat, curr_feat, most_common_angle, weight, curr_frame)
            self.i = self.i + 1

        self.pans = self.pan_counter.movements
        self.tilts = self.tilt_counter.movements

        self.clear()

    def clear(self):
        """
        This method is used to reset all internal counters and destroy all cv2 windows.
        """
        print("Clear and release everything. Reset frame index.")

        if self.mode > Runmodi.DEBUG_MODE:
            self.clear_outputs()

        if self.pan_counter.movement_created:
            self.pan_counter.movements[self.pan_counter.movements.__len__() - 1][1] = self.i + self.sf
        if self.tilt_counter.movement_created:
            self.tilt_counter.movements[self.tilt_counter.movements.__len__() - 1][1] = self.i + self.sf
        # reset i
        self.i = 1
        #self.cap.release()
        cv2.destroyAllWindows()

    def configure(self, config):
        """
        This method is used for reading configuration parameters (just needed during developing phase).

        :param config: This parameter must hold a valid dictionary including the mandatory parameters.
        """
        self.sensitivity = int(config["SENSITIVITY"])
        self.specificity = int(config["SPECIFICITY"])
        self.border = int(config["BORDER"])
        self.number_of_features = int(config["NUMBER_OF_FEATURES"])
        self.angle_diff_limit = int(config["ANGLE_DIFF_LIMIT"])
        self.mode = Runmodi(int(config["MODE"]))
        self.fpath = "/".join([config["INPUT_PATH"], config["INPUT_VIDEO"]])
        self.sf = int(config["BEGIN_FRAME"])
        self.ef = int(config["END_FRAME"])

    def __str__(self):
        """
        This method is used to create a summary string including key information related to the optical flow process.
        :return: This method returns a string summary of internal calculated key information.
        """
        return "\n".join(
            [
                "OFCMClassifier:",
                "- Video path: " + self.fpath,
                "- first frame to be captured: " + str(self.sf),
                "- last frame to be captured: " + str(self.ef),
                "Parameters: ",
                "- sensitivity: " + str(self.sensitivity) + " consecutive frames"
                "- border: " + str(self.border) + " for random features"
                "- number of features: " + str(self.number_of_features) + " to be tracked",
                "- angle diff limit: " + str(self.angle_diff_limit) + " degrees for background consideration.",
                "running mode " + self.mode.name
            ]
        )

    def re_init_ef(self, ef):
        """
        This method is used to re-initialize the end frame index.

        :param ef: This parameter must hold a valid frame index.
        """
        self.ef = ef
        self.end = self.ef - self.sf
        self.most_common_angles = np.zeros([self.end - 1, 1])
        self.weights = np.zeros([self.end - 1, 1])

        self.pan_counter = ClassificationCounter(np.zeros([self.end, 2]).astype(int))
        self.tilt_counter = ClassificationCounter(np.zeros([self.end, 2]).astype(int))

    def annotate_pan(self):
        """
        This method is used to annotate a frame as PAN.
        """
        if self.tilt_counter.movement_created:
            self.tilt_counter.end_movement(self.i + self.sf)
        if self.pan_counter.movement_created:
            self.pan_counter.end_movement(self.i + self.sf)
        else:
            self.pan_counter.add_movement(self.i + self.sf)

    def annotate_tilt(self):
        """
        This method is used to annotate a frame as TILT.
        """

        if self.pan_counter.movement_created:
            self.pan_counter.end_movement(self.i + self.sf)
        if self.tilt_counter.movement_created:
            self.tilt_counter.end_movement(self.i + self.sf)
        else:
            self.tilt_counter.add_movement(self.i + self.sf)

    def annotate_nomove(self):
        """
        This method is used to annotate a frame as NA.
        """

        if self.pan_counter.movement_created:
            self.pan_counter.end_movement(self.i + self.sf)
        if self.tilt_counter.movement_created:
            self.tilt_counter.end_movement(self.i + self.sf)

    def run_manual_evaluation(self):
        """
        This method is used to run optical flow process in DEBUG mode. A valid X-Server is needed to visualize the frame player.
        """
        self.cap = cv2.VideoCapture(self.fpath)

        if self.ef == 1:
            self.re_init_ef(int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT)))
        self.cap.set(1, self.sf)

        if self.mode > Runmodi.DEBUG_MODE:
            print("\n".join(["Video path: " + self.fpath,
                            "first frame to be captured: " + str(self.sf),
                            "last frame to be captured: " + str(self.ef)]))

        ret, curr_frame = self.cap.read()

        if not ret:
            raise Exception("Failing to open video file.")

        h, w, _ = curr_frame.shape
        self.frame_size = (w, h)

        print("configuring outputs.")
        self.out = self.init_outputs(self.fpath, self.sf, self.ef, (w, h))

        while self.i < self.end and self.cap.isOpened():

            ret, curr_frame = self.cap.read()

            if not ret:
                raise Exception("Failing to retrieve frame.")

            img = curr_frame.copy()
            img2 = curr_frame.copy()
            h, w, d = curr_frame.shape
            self.add_text(img, self.str_movement_info_manually(), h)
            self.add_text(img2, self.str_movement_info_manually(), int(h/2))
            title = str("Manual annotation")
            cv2.namedWindow(title, cv2.WINDOW_NORMAL)
            f = self.frame_size[1] / self.frame_size[0]
            cv2.resizeWindow(title, 500, np.round(500 * f).astype(int))
            cv2.imshow(title, img2)
            self.key_handler.wait_and_handle_key()

            self.out["manually"].write(img)
            self.i = self.i + 1

        if self.pan_counter.movement_created:
            self.pan_counter.end_movement(self.i + self.sf)
        if self.tilt_counter.movement_created:
            self.tilt_counter.end_movement(self.i + self.sf)

        self.pans = self.pan_counter.movements
        self.tilts = self.tilt_counter.movements

        self.clear()

    @staticmethod
    def compute_output_path(name, sf, ef, info):
        return "_".join([
            name,
            str(sf),
            str(ef),
            info,
            ".avi"
        ])

    # size=(width, height) of frame
    def init_outputs(self, fpath, sf, ef, size):
        s = fpath.split('/')
        s = s[s.__len__() - 1].split('.')

        import os
        try:
            opath = self.debug_path + "/" + s[0]
            os.mkdir(opath)
            print("successfully created output path ", opath)
        except OSError:
            print("failed to create directory. saving to working directory.")

        out_original = cv2.VideoWriter(opath + "/" + self.compute_output_path(s[0], sf, ef, "original"), cv2.VideoWriter_fourcc(*"MJPG"), 20.0, size)
        out_of = cv2.VideoWriter(opath + "/" + self.compute_output_path(s[0], sf, ef, "opticalFlow"), cv2.VideoWriter_fourcc(*"MJPG"), 20.0, size)
        out_mca = cv2.VideoWriter(opath + "/" + self.compute_output_path(s[0], sf, ef, "mostCommonAngle"), cv2.VideoWriter_fourcc(*"MJPG"), 20.0, size)
        out_manually = cv2.VideoWriter(opath + "/" + self.compute_output_path(s[0], sf, ef, "mannually"), cv2.VideoWriter_fourcc(*"MJPG"), 20.0, size)

        return {"original": out_original, "of": out_of, "mca": out_mca, "manually" : out_manually}

    def clear_outputs(self):
        for k, v in self.out.items():
            print("Releasing output element: " + k)
            v.release()

    # creates n random features within specific range r
    @staticmethod
    def create_random_features(n, r):
        feat = np.random.randint(low=r[0], high=r[1], size=(n, 1, 2)).astype(np.float32)
        return feat

    # returns minimum of height and width of frame minus self.border
    def range_of_frame(self, frame):
        h, w, _ = frame.shape
        return np.array([self.border, min(h, w)])

    def str_step_info(self):
        return str(self.i + self.sf) + "/" + str(self.ef)

    def getORBMatches(self, frame1, frame2):
        kp_curr_list = []
        kp_prev_list = []

        orb = cv2.ORB_create()

        kp_prev, descriptor_prev = orb.detectAndCompute(frame1, None)
        kp_curr, descriptor_curr = orb.detectAndCompute(frame2, None)

        out_frame1 = frame1.copy()
        # img = cv2.drawKeypoints(old_frame, kp, out_img)
        # out_img_prev = cv2.drawKeypoints(frames_np[i-1],
        #                                 kp_prev,
        #                                 out_img_prev,
        #                                 flags=cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS)
        out_frame2 = frame2.copy()
        # img = cv2.drawKeypoints(old_frame, kp, out_img)
        # out_img_curr = cv2.drawKeypoints(frames_np[i - 1],
        #                                 kp_curr,
        #                                 out_img_curr,
        #                                 flags=cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS)
        # print(out_img_prev.shape)
        # print(out_img_curr.shape)

        # Create a Brute Force Matcher object.
        bf = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True)

        # Perform the matching between the ORB descriptors of the training image and the test image
        matches = bf.match(descriptor_prev, descriptor_curr)

        # The matches with shorter distance are the ones we want.
        matches = sorted(matches, key=lambda x: x.distance)

        for match in matches:
            kp_curr_list.append(kp_curr[match.trainIdx].pt)
            kp_prev_list.append(kp_prev[match.queryIdx].pt)

        '''
        result = cv2.drawMatches(out_img_curr, kp_curr, out_img_prev, kp_prev, matches[:1], out_img_prev, flags=2)
        # Display the best matching points
        plt.title('Best Matching Points')
        plt.imshow(result)
        plt.draw()
        plt.pause(0.05)
        '''
        return np.array(kp_prev_list).astype(np.float32).reshape(-1, 1, 2), np.array(kp_curr_list).astype(np.float32).reshape(-1, 1, 2)

    # Store previous image, I_{i-1} and get current image I_i
    def step(self, curr_frame, curr_feat):
        print(self.i)

        print(curr_feat)
        if curr_feat == None:
            curr_feat = []

        prev_frame = curr_frame
        curr_frame = self.video_frames[self.i]

        print("Update step: frame " + self.str_step_info())

        prev_feat = curr_feat

        if curr_feat.__len__() < self.number_of_features:
            print("Not enough features. adding " +
                  str(self.number_of_features - curr_feat.__len__()) +
                  " new features.")

            # TODO: replace with meaningful feature extractor
            prev_feat = self.create_random_features(self.number_of_features, self.range_of_frame(curr_frame))

            '''
            #prev_feat, curr_feat = self.getORBMatches(curr_frame_gray, curr_frame_gray)

            # params for ShiTomasi corner detection
            feature_params = dict(maxCorners=100,
                                  qualityLevel=0.0003,
                                  minDistance=7,
                                  blockSize=7
                                  )

            print(curr_frame.shape)
            curr_frame_gray = cv2.cvtColor(curr_frame, cv2.COLOR_BGR2GRAY)
            prev_feat = cv2.goodFeaturesToTrack(curr_frame_gray, mask=None, **feature_params)
            print(type(prev_feat))
            '''
            ''''''

            #if(prev_feat == None):
            #    prev_feat = self.create_random_features(self.number_of_features, self.range_of_frame(curr_frame))
            print(len(curr_feat))
            print(len(prev_feat))


            prev_feat[0:curr_feat.__len__()] = curr_feat[0:prev_feat.__len__()]

        if self.mode > Runmodi.DEBUG_MODE:
            self.out["original"].write(curr_frame)

        return prev_frame, prev_feat, curr_frame, curr_feat

        # Store previous image, I_{i-1} and get current image I_i

    def step_eval(self, curr_frame, curr_feat):
        print(self.i)

        prev_frame = curr_frame
        ret, curr_frame = self.cap.read()

        print("Update step: frame " + self.str_step_info())

        prev_feat = curr_feat
        if curr_feat.__len__() < self.number_of_features:
            print("Not enough features. adding " +
                  str(self.number_of_features - curr_feat.__len__()) +
                  " new features.")
            prev_feat = self.create_random_features(self.number_of_features, self.range_of_frame(curr_frame))
            prev_feat[0:curr_feat.__len__()] = curr_feat

        if self.mode > Runmodi.DEBUG_MODE:
            self.out["original"].write(curr_frame)

        return prev_frame, prev_feat, curr_frame, curr_feat

    def optical_flow(self, prev_frame, prev_feat, curr_frame):
        """
        This method is used to calculate the optical flow between two consecutive frames.

        :param prev_frame: This parameter must hold a valid numpy frame (previous).
        :param prev_feat:  This parameter must hold valid features of the previous frame.
        :param curr_frame: This parameter must hold a valid numpy frame (current).
        :return: This method returns two arrays including the features of the previous frame as well as of the current frame.
        """
        print("Calculating optical flow: frame " + self.str_step_info())
        lk_params = dict(winSize=(15, 15),
                         maxLevel=2,
                         criteria=(cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03))

        #print(prev_feat)

        curr_feat, st, err = cv2.calcOpticalFlowPyrLK(cv2.cvtColor(prev_frame, cv2.COLOR_BGR2GRAY),
                                                      cv2.cvtColor(curr_frame, cv2.COLOR_BGR2GRAY), prev_feat, None,
                                                      **lk_params)

        # Select good points
        curr_feat = curr_feat[st == 1].reshape(-1, 1, 2)
        prev_feat = prev_feat[st == 1].reshape(-1, 1, 2)

        if self.mode > Runmodi.NORMAL_MODE:
            # debug or saving mode.
            img = curr_frame.copy()
            for i in range(curr_feat.__len__()):
                img = cv2.arrowedLine(img,(prev_feat[i, 0, 0], prev_feat[i, 0, 1]),
                                      (curr_feat[i, 0, 0], curr_feat[i, 0, 1]), [0, 0, 0], 2, 0)

            if self.mode > Runmodi.DEBUG_MODE:
                # saving mode
                self.out["of"].write(img)
            if self.mode == Runmodi.DEBUG_MODE or self.mode == Runmodi.DEBUG_AND_SAVE_MODE:
                # is debug mode too

                f = self.frame_size[1] / self.frame_size[0]

                title = str("Total Optical Flow per Frame.")
                cv2.namedWindow(title, cv2.WINDOW_NORMAL)
                cv2.resizeWindow(title, 500, np.round(500 * f).astype(int))
                cv2.imshow(title, img)
                self.key_handler.wait_and_handle_key()

        return prev_feat, curr_feat

    def compute_magnitude_angle(self, prev_feat, curr_feat):
        """
        This method is used to calculate the magnitude and angle between two consecutive feature frames.

        :param prev_feat: This parameter must hold a valid feature vector.
        :param curr_feat: This parameter must hold a valid feature vector.
        :return: This method returns the magnitude and angle between the given feature vectors.
        """
        print("Determine magnitude and angle (degrees) of features: frame " + self.str_step_info())

        if prev_feat.__len__() <= 0:
            print("no previous features... returning")
            assert(prev_feat.__len__() > 0)
        if prev_feat.__len__() != curr_feat.__len__():
            print("length is not correct")
            assert (prev_feat.__len__() == curr_feat.__len__())
        d = curr_feat - prev_feat
        mag = np.hypot(d[:, 0, 0], d[:, 0, 1])
        ang = np.round(np.degrees(np.arctan2(d[:, 0, 1], d[:, 0, 0])))

        return mag, ang

    def str_movement_info(self):
        """
        This method is used to visualize camera movement information.

        :return: This method returns a string including internal movement information.
        """
        return "PAN" if self.pan_counter.oversteps(1, self.sensitivity) else \
            ("TILT" if self.tilt_counter.oversteps(1, self.sensitivity) else "NO MOVEMENT")

    def str_movement_info_manually(self):
        """
        This method is used to visualize camera movement information (for manual debugging mode only).

        :return: This method returns a string including internal movement information.
        """
        return "PAN" if self.pan_counter.movement_created else \
            ("TILT" if self.tilt_counter.movement_created else "NO MOVEMENT")

    def add_text(self, img, text, h):
        """
        This method is used to add text to a given frame.

        :param img: This parameter must hold a valid numpy image.
        :param text: This parameter must hold a valid text string.
        :param h: This parameter must hold an integer which represents the text height.
        """
        font = cv2.FONT_HERSHEY_SIMPLEX
        bottom_left_corner = (20, h - 20)
        font_scale = 1
        font_color = (0, 0, 255)
        line_type = 2

        cv2.putText(img, text,
                    bottom_left_corner,
                    font,
                    font_scale,
                    font_color,
                    line_type)

    def estimate_background(self, prev_feat, curr_feat, most_common_angle, weight, curr_frame=None):
        """
        This method is used to estimate and distinguish between foreground and background objects.

        :param prev_feat: This parameter must hold a valid feature vector.
        :param curr_feat: This parameter must hold a valid feature vector.
        :param most_common_angle: This parameter must hold a valid integer representing the most common angle.
        :param weight: This parameter must hold a weight factor [0,1].
        :param curr_frame: This parameter must hold a valid numpy image.
        :return: This method returns the calculated most_common_angle, background_feat and new weight factor.
        """
        print("Convert relative pan / tilt units and estimate background pixel movement.")

        mag, ang = self.compute_magnitude_angle(prev_feat, curr_feat)
        count = np.histogram(ang, bins=np.arange(361))
        background_feat = curr_feat

        if most_common_angle is not None:
            print("\n".join([
                "Previous most common angle of entire optical flow is " + str(most_common_angle) + " degrees.",
                "Removing flow vectors whose angle differs more than " + str(self.angle_diff_limit) +
                " from it. Those are considered to be part of moving elements."
            ]))

            background = np.abs(ang - most_common_angle) < self.angle_diff_limit
            background_feat = curr_feat[background]
            moving_feat = curr_feat[~background]

            prev_background_feat = prev_feat[background]
            prev_moving_feat = prev_feat[~background]

            if self.mode > Runmodi.NORMAL_MODE and curr_frame is not None:
                img = curr_frame.copy()

                for f, f2 in zip(background_feat, prev_background_feat):
                    img = cv2.arrowedLine(img, (f2[0, 0], f2[0, 1]), (f[0, 0], f[0, 1]),
                                          [255, 0, 0], 2, 0)
                for m, m2 in zip(moving_feat, prev_moving_feat):
                    img = cv2.arrowedLine(img, (m2[0, 0], m2[0, 1]), (m[0, 0], m[0, 1]), [0, 255, 0], 2, 0)

                h, w, d = curr_frame.shape
                self.add_text(img, self.str_movement_info(), h)

                p1 = np.array([w / 2, h / 2]).reshape([2, 1])

                avg_mag = np.average(mag[background])
                if np.isnan(avg_mag):
                    avg_mag = 1
                rad = np.radians(most_common_angle)
                fac = weight * np.minimum(w, h)  # * avg_mag / 200
                p2 = p1 + np.array([fac * np.cos(rad), fac * np.sin(rad)])
                img = cv2.arrowedLine(img, (int(p1[0]), int(p1[1])), (int(p2[0]), int(p2[1])), [0, 0, 255], 2, 0)

                if self.mode > Runmodi.DEBUG_MODE:
                    # saving to output file
                    self.out["mca"].write(img)
                if self.mode == Runmodi.DEBUG_MODE or self.mode == Runmodi.DEBUG_AND_SAVE_MODE:
                    # is debug mode too
                    title=str("Background / Moving extracted with Most common angle.")
                    f = self.frame_size[1] / self.frame_size[0]
                    cv2.namedWindow(title, cv2.WINDOW_NORMAL)
                    cv2.resizeWindow(title, 500, np.round(500 * f).astype(int))
                    cv2.imshow(title, img)
                    self.key_handler.wait_and_handle_key()

        else:
            print("No previous most common angle. No background estimation.")

        max_idx = np.argmax(count[0])
        amount = count[0][max_idx]
        weight = amount / curr_feat.__len__()

        most_common_angle = count[1][max_idx]

        print("\n".join(["New most common angle " + str(most_common_angle) + " degrees.",
                         "- appears " + str(amount) + " times.",
                         "- remaining " + str(background_feat.__len__()) + " background features.",
                         "- weight: " + str(weight)]))

        return most_common_angle, background_feat, weight

    @staticmethod
    def contributes_to(angle, classifier, i, counter, offset):
        if classifier.classify(angle):
            counter.inc(1)
            counter.fill_totals(i, 1, offset)
            return True
        counter.inc(0)
        counter.fill_totals(i, 0, offset)
        return False

    def check_for_movement(self, angle, counter, classifier):
        contributes_to_movement = self.contributes_to(angle, classifier, self.i, counter, self.sf)

        if contributes_to_movement:
            if counter.equals(1, self.sensitivity):
                print("\n".join(["Last " + str(self.sensitivity) + " frames where " + str(classifier) + ".",
                                 "This is a " + str(classifier) + ".",
                                 "Resetting negative counter."]))
                counter.add_movement(self.i + self.sf - self.sensitivity)

            counter.reset(0)
        elif not contributes_to_movement:
            if counter.equals(0, self.specificity):
                print("\n".join(["Last " + str(self.specificity) + " frames where no " + str(classifier) + ".",
                                 "Resetting positive counter."]))
                counter.reset(1)
                counter.end_movement(self.i + self.sf)

    def interrupt(self):
        print("Interrupting.")
        self.end = self.i + 1

        self.most_common_angles = np.resize(self.most_common_angles, [self.end, 1])
        self.weights = np.resize(self.weights, [self.end, 1])
        self.pan_counter.resize(self.end, self.sf + self.i)
        self.tilt_counter.resize(self.end, self.sf + self.i)

    def init_savemode(self):
        if self.mode > Runmodi.DEBUG_MODE:
            print("Already in save mode. No action.")
            return
        print("Enabling save mode.")
        self.init_outputs(self.fpath, self.i + self.sf, self.ef, self.frame_size)
        self.mode = Runmodi.SAVE_MODE

    def init_debugmode(self):
        if self.mode > Runmodi.DEBUG_MODE:
            print("Decreasing mode. Clearing outputs.")
            self.clear_outputs()
        print("Enabling debug mode.")
        self.mode = Runmodi.DEBUG_MODE

    def init_savedebugmode(self):
        if self.mode <= Runmodi.DEBUG_MODE:
            print("Initializing outputs.")
            self.init_outputs(self.fpath, self.i + self.sf, self.ef, self.frame_size)
        print("Enabling save debug mode.")
        self.mode = Runmodi.DEBUG_AND_SAVE_MODE

    def init_normalmode(self):
        if self.mode > Runmodi.DEBUG_MODE:
            print("Decreasing mode. Clearing outputs.")
            self.clear_outputs()
        print("Enabling normal mode.")
        self.mode = Runmodi.NORMAL_MODE

    def plot_movements(self, show=True):
        fig = plt.figure()

        x = np.arange(self.sf, self.sf + self.end, 1)
        y = np.resize(self.most_common_angles, x.shape)

        xy = np.array([x, y])
        plt.scatter(x, y, color='lightgrey')
        # pans = xy[:, self.pan_counter.totals[:, 1].astype(bool)]
        # plt.scatter(pans[0], pans[1], color='red')
        # tilts = xy[:, self.tilt_counter.totals[:, 1].astype(bool)]
        # plt.scatter(tilts[0], tilts[1], color='blue')

        self.pan_counter.draw_movement(xy, self.sf + 1, 'red', 'darkred', 'pan')
        self.tilt_counter.draw_movement(xy, self.sf + 1, 'blue', 'darkblue', 'tilt')

        plt.xlabel('Frame Index')
        plt.ylabel('Angle Degrees')
        plt.ylim([0, 180])
        plt.title('Classified Frames')

        plt.legend()
        if show:
            plt.show()

    def to_avi(self, opath):
        print("saving annotated video to ", opath, " ...")
        self.cap = cv2.VideoCapture(self.fpath)
        self.cap.set(1, self.sf)
        ret, curr_frame = self.cap.read()

        if not ret:
            raise Exception("Failing to open video file.")

        h, w, _ = curr_frame.shape
        self.frame_size = (w, h)
        out = cv2.VideoWriter(opath, cv2.VideoWriter_fourcc(*"MJPG"), 20.0, self.frame_size)

        while self.i < self.end and self.cap.isOpened():

            img = curr_frame.copy()
            if self.pan_counter.is_movement(self.i + self.sf):
                self.add_text(img, "PAN", self.frame_size[1])
            if self.tilt_counter.is_movement(self.i + self.sf):
                self.add_text(img, "TILT", self.frame_size[1])

            p1 = np.array([w / 2, h / 2]).reshape([2, 1])
            rad = np.radians(self.most_common_angles[self.i - 1])
            fac = 0.5 * np.minimum(w, h)# * avg_mag / 200
            p2 = p1 + np.array([fac * np.cos(rad), fac * np.sin(rad)])
            img = cv2.arrowedLine(img, (int(p1[0]), int(p1[1])), (int(p2[0]), int(p2[1])), [0, 0, 255], 2, 0)
            out.write(img)
            if self.mode == Runmodi.DEBUG_MODE or self.mode == Runmodi.DEBUG_AND_SAVE_MODE:
                cv2.imshow("saving ...", img)
                self.key_handler.wait_and_handle_key()

            self.i = self.i + 1
            ret, curr_frame = self.cap.read()

        out.release()
        self.i = 1
        self.cap.release()
        if self.mode == Runmodi.DEBUG_MODE or self.mode == Runmodi.DEBUG_AND_SAVE_MODE:
            cv2.destroyAllWindows()
        print("saving annotated video to ", opath, " DONE.")

    def to_png(self, opath):
        print("saving movements plot to ", opath, " ...")
        self.plot_movements(False)
        plt.savefig(opath)
        print("saving movements plot to ", opath, " DONE.")

    def to_csv(self, opath):
        print("saving extracted movements to ", opath, " ...")
        import pandas
        # add all pans
        movements = []
        for (sf, ef) in self.pan_counter.movements:
            movements.append([self.fpath, 'PAN', sf, ef])
        for (sf, ef) in self.tilt_counter.movements:
            movements.append([self.fpath, 'TILT', sf, ef])
        movements = np.array(movements)
        if movements.__len__() == 0:
            print("no movements detected to save to csv.")
            return
        df = pandas.DataFrame({'name': movements[:, 0], 'movement': movements[:, 1], 'start': movements[:, 2], 'end': movements[:, 3]})
        df.to_csv(opath, mode='a', header=False, index=False, sep=";") # header must be there at beginning!!
        print("saving extracted movements to ", opath, " DONE.")